require 'timeout'
require 'socket'
require 'resolv'

module IPAM
  class PingRandomDb < Base
    def generator
      @generator ||= Random.new(mac ? mac.delete(':').to_i(16) : Random.new_seed)
    end

    def random_ip(range)
      IPAddr.new(range.sample(random: generator), subnet.family)
    end

    def tcp_pingable?(ip)

      ports = [7, 22, 80, 443]

      begin
        ports.each do |port|
          if tcp_pingable_by_port?(ip, port)
            return true
          end
        end
        false
      end

    end

    def tcp_pingable_by_port?(ip, port)
      # This code is from net-ping, and stripped down for use here

      # Whether or not Errno::ECONNREFUSED is considered a successful ping
      @service_check = true
      @timeout = 1
      bool = false

      begin
        bool = Socket.tcp(ip, port, connect_timeout: @timeout) { true }
      rescue Errno::ECONNREFUSED
        if @service_check
          bool = true
        end
      rescue Exception
        bool = false
      end

      if bool
        logger.info "[IPAM] Succesful telnet ping #{ip}, port #{port}"
      end

      bool
    rescue
      true
    end

    def icmp_pingable? ip
      # Always shell to ping, instead of using net-ping
      if RUBY_PLATFORM =~ /mingw/
        # Windows uses different options for ping and does not have /dev/null
        system("ping -n 2 -w 1000 #{ip} > NUL")
      else
        # Default to Linux ping options and send to /dev/null
        system("ping -c 2 -W 1 #{ip} > /dev/null")
      end
    rescue => err
      logger.warn "[IPAM] Unable to icmp ping #{ip} because #{err.inspect}."
      true
    end

    def ns_resolvable? ip
      begin
        name = Resolv.getname ip
        logger.warn "[IPAM] Found a DNS resolvable IP #{ip}, resolved name: #{name}."
        true
      rescue StandardError
        false
      end
    end

    # Safety check not to spend much CPU time when there are no many free IPs left. This gives up
    # in about a second on Ryzen 1700 running with Ruby 2.4.
    MAX_ITERATIONS = 100_000

    def suggest_ip
      iterations = 0
      # Remove IPs already excluded or known.
      range = subnet_range.to_a - excluded_ips.to_a
      range -= subnet.known_ips.to_a

      unless range.empty?
        logger.info "[IPAM] IP range from #{range.first.to_s} to #{range.last.to_s}, total: #{range.length}"
      end

      loop do
        if range.empty?
          errors.add(:subnet, _('no free IP could be found in our DB, enlarge subnet range'))
          return nil
        end

        candidate = random_ip(range)
        iterations += 1
        break if iterations >= MAX_ITERATIONS
        # try to match it
        ip = candidate.to_s
        logger.info "[IPAM] Searching for free IP - resolving #{ip}"

        # Check again if something has been changed.
        if !excluded_ips.include?(ip) && !subnet.known_ips.include?(ip)
          if tcp_pingable?(ip) || icmp_pingable?(ip)
            logger.warn("[IPAM] Found a pingable IP (#{ip}) address which not marked as known. Skipping it...")
          else
            unless ns_resolvable?(ip)
              logger.info("[IPAM] Found #{ip} in #{iterations} iterations")
              return ip
            end
          end
        end

        range -= [candidate]
      end

      logger.warn("[IPAM] Not suggesting IP Address for #{subnet} as no free IP found in reasonable time (#{iterations} iterations)")
      errors.add(:subnet, _('no random free IP could be found in our DB, enlarge subnet range'))
      nil
    end
  end
end
